# Short JSON Schema

Short JSON Schema is an abbreviated syntax for a subset of [JSON Schema][jsonschema].  

JSON Schema is a powerful language for constraining the format of JSON structures, but its standard, JSON-based syntax is very verbose.  

For example, take the following JSON document:

```json
{
    "numberProperty": 123,
    "stringProperty": "abc"
}
```

A schema for this in [JSON Schema][jsonschema] might look something like this:

```json
{
    "type": "object",
    "properties": {
        "numberProperty": {
            "type": "number"
        },
        "stringProperty": {
            "type": "string"
        }
    },
    "required": [
        "numberProperty",
        "stringProperty"
    ],
    "additionalProperties": false
}
```

In Short JSON Schema, this becomes:

```
{
    numberProperty:#,
    stringProperty:$
}
```

[jsonschema]: http://json-schema.org/

## Motivation

When writing code that consumes JSON, you should be defensive against JSON that
doesn't match your expectations.  Most languages make it easy to defend against
properties that may or may not be present.  For example:

```python
someJSON.get("thing", {}).get("subThing", None)
if subThing is not None:
   ... # "thing" or "subThing" not present
```

What is less easy is guarding against properties being present with the wrong type.  For example, if `thing` is a list:

```python
json.loads('{"thing": [ ] }').get("thing", {}).get("subThing", None)
```

We get:

```
AttributeError: 'list' object has no attribute 'get'
```

JSON Schema is a great way to specify declaratively the structure that you
expect, but in its standard syntax, it's often more verbose and barely more
readable than doing type checks manually in code.

Short JSON Schema is intended to allow concise specification of JSON formats,
so that your code needn't be cluttered with checks for the correct structure.

## Scope

Short JSON Schema supports a subset of JSON Schema functionality.  The
`shortjsonschema` Python module converts Short JSON Schema syntax into standard
JSON Schema.


## Examples

An object with a single, mandatory string property:

```
{
  thing:$
}
```

e.g.

```javascript
{
    "thing": "this is valid"
}
```

### Scalar types

Other JSON types are supported with `#` for number, `B` for boolean, and `~`
for null.  Multiple types can be specified for a single property.

```
{
  optionalNumberProperty?:#,
  booleanProperty:B,
  nullableString:$~
}
```

### Objects

```
{
   thing:{              -- comments after '--'
       subThing1:$,  
       subThing2:#,
       ...              -- additionalProperties = True (default is False)
   }
}
```

### Arrays

```
{
    arrayOfNumbers: [#],
    arrayOfStringsOrNumbers:[$#]
}
```

### Formats

```
{
    dateProperty:$(date)
}

## Defaults

In Short JSON Schema, properties are required by default, and additional
properties are disallowed.  Properties can be made optional by adding a `?`
modifier, and additional properties can be allowed by adding `...` in the
property list.

```
{
   myString?:$,
   ...
}
```

All properties allowed, but if present, `myString` must be a string.


