from shortjsonschema import load
import jsonschema
import pytest

def test_short_schema():
    tests = [
        # Single required property, whitespace everywhere
        { 
            "input": " { requiredString : $ } ",
            "output": { 
                "type": "object",
                "properties": {
                    "requiredString": {
                        "type": "string",
                    }
                },
                "required": ["requiredString"],
                "additionalProperties": False
            },
            "valid_instances": [
                { "requiredString": "abcd" },
                { "requiredString": "" },
            ],
            "invalid_instances": [
                { "requiredString": 123 },
                {  },
                { "requiredString": None },
            ],

        },
        # Multi-line, mandatory and optional properties, trailing comma
        { 
            "input": """
            {
                thing?:{
                    subthing1?:$,  -- with comment
                    subthing2:$,
                }
            }
            """,
            "output": { 
                "type": "object",
                "properties": {
                    "thing": {
                        "type": "object",
                        "properties": {
                            "subthing1": { "type": "string" },
                            "subthing2": { "type": "string" },
                        },
                        "required": ["subthing2"],
                        "additionalProperties": False
                    }
                },
                "additionalProperties": False
            },
            "valid_instances": [
                { "thing": { "subthing1": "foo", "subthing2": "bar" } },
                { "thing": { "subthing2": "bar" } },
            ],
            "invalid_instances": [
                { "thing": { "subthing1": "foo", "subthing2": "bar", "subthing3": "bar" } },
                { "thing": { "subthing1": "foo" } },
                { "thing": [] },
                { "thing": None },
            ]
        },
        # Additional properties allowed
        { 
            "input": "{thing?:$#, ...}",
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { "type": ["string", "number"] }
                },
                "additionalProperties": True
            },
            "valid_instances": [
                { "thing": 123, "otherThing": "abc", "otherOtherThing": {} },
                { "thing": "string", "otherThing": "abc", "otherOtherThing": {} },
            ],
            "invalid_instances": [
                { "thing": True, "otherThing": "abc", "otherOtherThing": {} },
            ],
        },
        { 
            "input": "{thing?:[$#]}",
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { "type": "array", "items": { "type": ["string", "number"] } }
                },
                "additionalProperties": False
            }
        },
        { 
            "input": "{thing:[B]}",
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { "type": "array", "items": { "type": "boolean" } },
                },
                "required": [ "thing" ],
                "additionalProperties": False
            }
        },
        { 
            "input": """
            {
                    thing:[ B$ ] }""",
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { "type": "array", "items": { "type": [ "boolean", "string"] } },
                },
                "required": [ "thing" ],
                "additionalProperties": False
            },
            "valid_instances": [
                { "thing": [ True, False ] },
                { "thing": [ "foo" ] },
                { "thing": [ "bar", False ] },
            ],
            "invalid_instances": [
                { "thing": [ 1 ] },
                { "thing": [ None ] },
                { "thing": { } },
            ],
        },
        # Both string and object
        { 
            "input": "{thing?:${otherThing?:$}}",
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { 
                        "type": ["string", "object" ], 
                        "properties": { "otherThing": { "type": "string" } } ,
                        "additionalProperties": False,
                    },
                },
                "additionalProperties": False,
            },
            "valid_instances": [
                { "thing": "baz" },
                { "thing": { "otherThing": "bar" } },
            ],
            "invalid_instances": [
                { "thing": [ ] },
            ]
        },
        # Empty object definitions are OK
        { 
            "input": "{thing?:{}}",
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { 
                        "type": "object",
                        "additionalProperties": False,
                    },
                },
                "additionalProperties": False,
            },
            "valid_instances": [
                { "thing": {} },
            ],
            "invalid_instances": [
            ]
        },

        # Base types
        { 
            "input": """
                {   
                    thing(BaseType)?,
                    otherThing(BaseType)?
                } 
                {
                    BaseType:# 
                }
            """,
            "output": { 
                "type": "object",
                "properties": {
                    "thing": { 
                        "$ref": "#/definitions/BaseType",
                    },
                    "otherThing": { 
                        "$ref": "#/definitions/BaseType",
                    },
                },
                "additionalProperties": False,
                "definitions": {
                    "BaseType": { "type": "number" }
                }
            },
            "valid_instances": [
                { "thing": 123 },
            ],
            "invalid_instances": [
                { "thing": "abc" },
            ]
        },

        # pattern properties
        { 
            "input": "{/^a.*$/:$}",
            "output": { 
                "type": "object",
                "patternProperties": {
                    "^a.*$": { "type": "string" },
                },
                "additionalProperties": False,
            },
            "valid_instances": [
                { "athing": "baz" },
                { "anotherThing": "otherThing" } ,
            ],
            "invalid_instances": [
                { "thing": "abc" },
                { "athing": 123 },
            ]
        },

        # different types
        { 
            "input": """
            {
                nullThing?:~, -- nullable
                booleanThing?:B, -- boolean
                dateThing?:$(date)
            }
            """,
            "output": { 
                "type": "object",
                "properties": {
                    "nullThing": {
                        "type": "null",
                    },
                    "booleanThing": {
                        "type": "boolean",
                    },
                    "dateThing": {
                        "type": "string",
                        "format": "date"
                    }
                },
                "additionalProperties": False
            },
            "valid_instances": [
                { "nullThing": None, "booleanThing": True, "dateThing": "2019-01-01" },
            ],
            "invalid_instances": [
                { "nullThing": False, "booleanThing": True },
                { "nullThing": False, "booleanThing": None },
                { "nullThing": None, "booleanThing": True, "dateThing": "abc" },
            ]
        },

        # Name quoting
        { 
            "input": """
            {
                "this is a key":$,
                "this is a \\"key\\""?:$
            }
            """,
            "output": { 
                "type": "object",
                "properties": {
                    "this is a key": {
                        "type": "string",
                    },
                    "this is a \"key\"": {
                        "type": "string",
                    },
                },
                "required": [ "this is a key" ],
                "additionalProperties": False
            },
            "valid_instances": [
                { "this is a key": "bas" },
                { "this is a key": "bas", 'this is a "key"': "blort" },
            ],
            "invalid_instances": [
                {},
            ]
        },

        # Patterns 
        { 
            "input": """
            {
                "p":$/f.o/#,
            }
            """,
            "output": { 
                "type": "object",
                "properties": {
                    "p": {
                        "type": ["string", "number"],
                        "pattern": "f.o"
                    },
                },
                "required": [ "p" ],
                "additionalProperties": False
            },
            "valid_instances": [
                { "p": "foo" },
                { "p": "fxo" },
                { "p": 37 },
            ],
            "invalid_instances": [
                {
                  "p": "bar"
                },
            ]
        },
    ]

    for t in tests:
        s = load(t["input"]) 
        assert s == t["output"]
        try:
            jsonschema.Draft4Validator.check_schema(s)
        except jsonschema.SchemaError:
            pytest.fail("Schema error parsing" + repr(s))

        sv = jsonschema.Draft4Validator(s, format_checker = jsonschema.FormatChecker())
        for i in t.get("valid_instances",{}):
            try:
                sv.validate(i) 
            except jsonschema.ValidationError:
                pytest.fail("error validating" + repr(i))

        for i in t.get("invalid_instances",{}):
            if sv.is_valid(i):
                pytest.fail("Instance passed validation unexpectedly: " + repr(i))


